/*
 * This is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This code is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this code; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.  */

/* Copyright (C) 2000 Ian Zimmerman <itz@speakeasy.org> */

static char rcsid[] = "$Id: connections.c,v 1.13 2002/06/07 15:01:16 itz Exp $";

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <stdlib.h>
#include <paths.h>
#include <unistd.h>
#include <fcntl.h>
#include <termios.h>
#include <search.h>
#include <assert.h>
#include <sys/select.h>

#include "connections.h"
#include "linuxonly.h"
#include "sockets.h"
#include "gim_event.h"
#include "gim_common.h"
#include "exc.h"

static struct gim_connect_t connections_by_fd[FD_SETSIZE];
static struct gim_connect_t connections_by_vt[MAX_NR_CONSOLES+1];

static void
do_init_connections (void)
{
  static int connections_by_vt_initialized;
  int i;

  if (!connections_by_vt_initialized) {
    for (i = 0; i <= MAX_NR_CONSOLES; i = i+1) {
      connections_by_vt[i].next = connections_by_vt[i].prev = &connections_by_vt[i];
    } /*for*/
    connections_by_vt_initialized = 1;
  } /*if*/
}

/* this assumes caller already checked there's no existing entry for
   the same fd */

void
add_connection (int fd)
{
  struct gim_connect_t* p;
  struct ucred sucr;

  p = &connections_by_fd[fd];
  assert(!p->allocated);
  p->state = gim_state_new;
  p->min_mod = 0;
  p->max_mod = ~0;
  p->next = p->prev = p;
  p->fd = fd;
  authenticate_client(fd, &sucr);
  p->owner = sucr.uid;
  p->pid = sucr.pid;
  p->allocated = 1;
}

void
set_connection_vt (struct gim_connect_t* p, int vt)
{
  do_init_connections();
  p->vt = vt;
                                /* insert into chain for searching by vt */
  p->next = connections_by_vt[vt].next;
  p->prev = &connections_by_vt[vt];
  connections_by_vt[vt].next->prev = connections_by_vt[vt].next = p;
}

void
unset_connection_vt (struct gim_connect_t* p)
{
  do_init_connections();
  p->prev->next = p->next;
  p->next->prev = p->prev;
  p->next = p->prev = p;
}

void
delete_connection (struct gim_connect_t* p)
{
  p->allocated = 0;
}

struct gim_connect_t*
get_connection_by_fd (int fd)
{
  struct gim_connect_t* p;

  p = &connections_by_fd[fd];
  assert(p->allocated);
  return p;
}

#define IS_SUBSET(a,b) (((a)|(b)) == (b))
                                /* return first enabled connection in vt chain */
struct gim_connect_t*
get_active_connection (int vt, unsigned mods)
{
  struct gim_connect_t* p;

  do_init_connections();
  p = connections_by_vt[vt].next;

  while (p != &connections_by_vt[vt]) {
    if (p->state == gim_state_open && !is_stopped(p->pid) &&
        IS_SUBSET(mods, p->max_mod) && IS_SUBSET(p->min_mod, mods)) {
      exc_ereturn2(struct gim_connect_t*, p);
    } /*if*/
    p = p->next;
  } /*while*/

  return 0;
}
